"""
本模块可用于与任何以太坊同构链（即通常所说的 EVM 链）进行交互，支持常用的链上交互操作。
"""

from web3 import Web3
from eth_account import Account as EthAccount
from loguru import logger
from typing import Optional, Union, List, Any
from traceback import format_exc
from json import dump, dumps
import os

LogPath = os.path.join("logs", "Poseidon_{time}.log")
logger.add(LogPath)


class Chain():
    """
    Chain 是区块链实例，后续的所有链上交互的操作都将经由该指定节点处理。
    """

    def __init__(self, RPCUrl: str, RequestParams: Optional[dict] = None):
        """
        初始化。根据给定的节点 RPC 地址进行连接，可通过代理访问。当连接节点失败时会抛出异常。

        参数：
            RPCUrl (str): 节点 RPC 地址
            RequestParams (Optional[dict]): 连接时使用的 request 参数，默认为 None。
            例如当需要使用代理进行访问时，则传入 RequestParams={"proxies": {"http": "http://localhost:<ProxyPort>","https": "http://localhost:<ProxyPort>"}}

        成员变量：
            ChainId (int): 链 ID
            Node (Web3.HTTPProvider): web3.py 原生的 HTTP 交互器实例
            Eth (Web3.HTTPProvider.eth): HTTP 交互器实例中的 eth 模块
        """

        from time import time
        from web3 import HTTPProvider
        from web3.middleware import geth_poa_middleware
        RequestParamsPrint = f"[RequestParams]{RequestParams}\n" if RequestParams else ""
        StartTime = time()
        self.Node = Web3(HTTPProvider(RPCUrl, request_kwargs=RequestParams))
        if self.Node.isConnected():
            FinishTime = time()
            Delay = round((FinishTime - StartTime) * 1000)
            logger.success(f"\n[Chain][Initialize]Connected to [{RPCUrl}] [{Delay} ms]\n{RequestParamsPrint}{'-'*80}")
            self.Node.middleware_onion.inject(geth_poa_middleware, layer=0)
            self.Eth = self.Node.eth
            self.GetBasicInformation()
        else:
            logger.error(f"\n[Chain][Initialize]Failed to connect to [{RPCUrl}]\n{RequestParamsPrint}{'-'*80}")
            raise Exception("Failed to connect to chain.")

    def GetBasicInformation(self) -> dict:
        """
        获取区块链基本信息。包括链 ID 、区块高度、 GasPrice 、出块间隔、当前节点的客户端软件版本号。

        返回值：
            BasicInformation (dict): 区块链基本信息构成的字典。
            {"ChainId"|"BlockNumber"|"GasPrice"|"Timeslot"|"ClientVersion"}
        """

        self.ChainId = self.Eth.chain_id
        BlockNumber = self.Eth.block_number
        GasPrice = self.Eth.gas_price
        Timeslot = self.Eth.get_block(BlockNumber).timestamp - self.Eth.get_block(BlockNumber - 1).timestamp
        ClientVersion = self.Node.clientVersion
        logger.success(
            f"\n[Chain][GetBasicInformation]\n[ChainId]{self.ChainId}\n[BlockNumber]{BlockNumber}\n[GasPrice]{Web3.fromWei(GasPrice, 'gwei')} Gwei\n[Timeslot]{Timeslot}s\n[ClientVersion]{ClientVersion}\n{'-'*80}"
        )
        return {"ChainId": self.ChainId, "BlockNumber": BlockNumber, "GasPrice": GasPrice, "Timeslot": Timeslot, "ClientVersion": ClientVersion}

    def GetTransactionInformationByHash(self, TransactionHash: str) -> dict:
        """
        根据交易哈希查询该交易的详细回执信息。包括交易哈希、所在区块号、交易索引号、交易状态、交易类型、交易行为、发送者、接收者、(部署的合约地址)、(GasPrice 或 (MaxFeePerGas 和 MaxPriorityFeePerGas))、GasLimit、GasUsed、Nonce、Value、R、S、V、Logs、InputData。

        参数：
            TransactionHash (str): 要查询的交易的哈希

        返回值：
            TransactionInformation (dict): 交易信息构成的字典。当出现异常时返回 None 。
            {"TransactionHash"|"BlockNumber"|"TransactionIndex"|"Status"|"Type"|"Action"|"From"|"To"|("ContractAddress")|<"GasPrice"|("MaxFeePerGas"&"MaxPriorityFeePerGas")>|"GasLimit"|"GasUsed"|"Nonce"|"Value"|"R"|"S"|"V"|"Logs"|"InputData"}
        """

        try:
            Info = self.Eth.wait_for_transaction_receipt(TransactionHash, timeout=90)
            BlockNumber = Info.blockNumber
            TransactionIndex = Info.transactionIndex
            Status = Info.status
            From = Info["from"]
            To = Info.to
            ContractAddress = Info.contractAddress
            GasUsed = Info.gasUsed
            Logs = Web3.toJSON(Info.logs)
            Info = self.Eth.get_transaction(TransactionHash)
            TransactionHash = Info.hash.hex()
            GasPrice = Info.gasPrice
            MaxFeePerGas = Info.get("maxFeePerGas", None)
            MaxPriorityFeePerGas = Info.get("maxPriorityFeePerGas", None)
            GasLimit = Info.gas
            Nonce = Info.nonce
            Value = Info.value
            R = Info.r.hex()
            S = Info.s.hex()
            V = Info.v
            InputData = Info.input
            Type = "EIP-1559" if MaxFeePerGas or MaxPriorityFeePerGas else "Traditional"
            Action = "Deploy Contract" if To == None else "Call Contract" if self.Eth.get_code(Web3.toChecksumAddress(To)).hex() != "0x" else "Normal Transfer"
            ContractPrint = f"[ContractAddress]{ContractAddress}\n" if ContractAddress else ""
            GasPricePrint = f"[GasPrice]{Web3.fromWei(GasPrice, 'gwei')} Gwei" if Type == "Traditional" else f"[MaxFeePerGas]{Web3.fromWei(MaxFeePerGas, 'gwei')} Gwei\n[MaxPriorityFeePerGas]{Web3.fromWei(MaxPriorityFeePerGas, 'gwei')} Gwei"
            GeneralPrint = f"\n[Chain][GetTransactionInformationByHash]\n[TransactionHash]{TransactionHash}\n[BlockNumber]{BlockNumber}\n[TransactionIndex]{TransactionIndex}\n[Status]{'Success' if Status else 'Fail'}\n[Type]{Type}\n[Action]{Action}\n[From]{From}\n[To]{To}\n{ContractPrint}{GasPricePrint}\n[GasLimit]{GasLimit} [GasUsed]{GasUsed}\n[Nonce]{Nonce} [Value]{Value}\n[R]{R}\n[S]{S}\n[V]{V}\n[Logs]{Logs}\n[InputData]{InputData}\n{'-'*80}"
            if Status:
                logger.success(GeneralPrint)
            else:
                logger.error(GeneralPrint)
            return {"TransactionHash": TransactionHash, "BlockNumber": BlockNumber, "TransactionIndex": TransactionIndex, "Status": Status, "Type": Type, "Action": Action, "From": From, "To": To, "ContractAddress": ContractAddress, "GasPrice": GasPrice, "MaxFeePerGas": MaxFeePerGas, "MaxPriorityFeePerGas": MaxPriorityFeePerGas, "GasLimit": GasLimit, "GasUsed": GasUsed, "Nonce": Nonce, "Value": Value, "R": R, "S": S, "V": V, "Logs": Logs, "InputData": InputData}
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Chain][GetTransactionInformationByHash]Failed\n[TransactionHash]{TransactionHash}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def GetTransactionInformationByBlockIdAndIndex(self, BlockID: Union[str, int], TransactionIndex: int) -> dict:
        """
        根据区块 ID 和交易在块中的索引来查询该交易的详细回执信息。包括交易哈希、所在区块号、交易索引号、交易状态、交易类型、交易行为、发送者、接收者、(部署的合约地址)、(GasPrice 或 (MaxFeePerGas 和 MaxPriorityFeePerGas))、GasLimit、GasUsed、Nonce、Value、R、S、V、Logs、InputData。

        参数：
            BlockID (Union[str,int]): 区块 ID 。可为区块号数值或 'latest', 'earliest', 'pending' 。
            TransactionIndex (int): 交易在块中的索引

        返回值：
            TransactionInformation (dict): 交易信息构成的字典。当出现异常时返回 None 。
            {"TransactionHash"|"BlockNumber"|"TransactionIndex"|"Status"|"Type"|"Action"|"From"|"To"|("ContractAddress")|<"GasPrice"|("MaxFeePerGas"&"MaxPriorityFeePerGas")>|"GasLimit"|"GasUsed"|"Nonce"|"Value"|"R"|"S"|"V"|"Logs"|"InputData"}
        """

        try:
            Info = self.Eth.get_transaction_by_block(BlockID, TransactionIndex)
            TransactionHash = Info.hash.hex()
            BlockNumber = Info.blockNumber
            TransactionIndex = Info.transactionIndex
            From = Info["from"]
            To = Info.to
            GasPrice = Info.gasPrice
            MaxFeePerGas = Info.get("maxFeePerGas", None)
            MaxPriorityFeePerGas = Info.get("maxPriorityFeePerGas", None)
            GasLimit = Info.gas
            Nonce = Info.nonce
            Value = Info.value
            R = Info.r.hex()
            S = Info.s.hex()
            V = Info.v
            InputData = Info.input
            Info = self.Eth.wait_for_transaction_receipt(TransactionHash, timeout=90)
            Status = Info.status
            GasUsed = Info.gasUsed
            ContractAddress = Info.contractAddress
            Logs = Info.logs
            Type = "EIP-1559" if MaxFeePerGas else "Traditional"
            Action = "Deploy Contract" if To == None else "Call Contract" if self.Eth.get_code(Web3.toChecksumAddress(To)).hex() != "0x" else "Normal Transfer"
            ContractPrint = f"[ContractAddress]{ContractAddress}\n" if ContractAddress else ""
            GasPricePrint = f"[GasPrice]{Web3.fromWei(GasPrice, 'gwei')} Gwei" if Type == "Traditional" else f"[MaxFeePerGas]{Web3.fromWei(MaxFeePerGas, 'gwei')} Gwei\n[MaxPriorityFeePerGas]{Web3.fromWei(MaxPriorityFeePerGas, 'gwei')} Gwei"
            GeneralPrint = f"\n[Chain][GetTransactionInformationByBlockIdAndIndex]\n[TransactionHash]{TransactionHash}\n[BlockNumber]{BlockNumber}\n[TransactionIndex]{TransactionIndex}\n[Status]{'Success' if Status else 'Fail'}\n[Type]{Type}\n[Action]{Action}\n[From]{From}\n[To]{To}\n{ContractPrint}{GasPricePrint}\n[GasLimit]{GasLimit} [GasUsed]{GasUsed}\n[Nonce]{Nonce} [Value]{Value}\n[R]{R}\n[S]{S}\n[V]{V}\n[Logs]{Logs}\n[InputData]{InputData}\n{'-'*80}"
            if Status:
                logger.success(GeneralPrint)
            else:
                logger.error(GeneralPrint)
            return {"TransactionHash": TransactionHash, "BlockNumber": BlockNumber, "TransactionIndex": TransactionIndex, "Status": Status, "Type": Type, "Action": Action, "From": From, "To": To, "ContractAddress": ContractAddress, "GasPrice": GasPrice, "MaxFeePerGas": MaxFeePerGas, "MaxPriorityFeePerGas": MaxPriorityFeePerGas, "GasLimit": GasLimit, "GasUsed": GasUsed, "Nonce": Nonce, "Value": Value, "R": R, "S": S, "V": V, "Logs": Logs, "InputData": InputData}
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Chain][GetTransactionInformationByBlockIdAndIndex]Failed\n[BlockID]{BlockID}\n[TransactionIndex]{TransactionIndex}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def GetBlockInformation(self, BlockID: Union[str, int]) -> dict:
        """
        根据区块 ID 获取该区块的详细信息。包括区块号、区块哈希、矿工、时间戳、GasLimit、GasUsed、块内交易的哈希集合。

        参数：
            BlockID (Union[str,int]): 区块 ID 。可为区块号数值或 'latest', 'earliest', 'pending' 。

        返回值：
            BlockInformation (dict): 区块信息构成的字典。当出现异常时返回 None 。
            {"BlockNumber"|"BlockHash"|"Miner"|"TimeStamp"|"GasLimit"|"GasUsed"|"Transactions"}     
        """

        try:
            Info = self.Eth.get_block(BlockID)
            BlockNumber = Info.number
            BlockHash = Info.hash.hex()
            Miner = Info.miner
            TimeStamp = Info.timestamp
            GasLimit = Info.gasLimit
            GasUsed = Info.gasUsed
            Transactions = Web3.toJSON(Info.transactions)
            logger.success(
                f"\n[Chain][GetBlockInformation]\n[BlockNumber]{BlockNumber}\n[BlockHash]{BlockHash}\n[Miner]{Miner}\n[TimeStamp]{TimeStamp}\n[GasLimit]{GasLimit}\n[GasUsed]{GasUsed}\n[Transactions]{Transactions}"
            )
            return {"BlockNumber": BlockNumber, "BlockHash": BlockHash, "Miner": Miner, "TimeStamp": TimeStamp, "GasLimit": GasLimit, "GasUsed": GasUsed, "Transactions": Transactions}
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Chain][GetBlockInformation]Failed\n[BlockID]{BlockID}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def GetBalance(self, Address: str) -> int:
        """
        根据账户地址获取其网络原生代币余额。

        参数：
            Address (str): 账户地址

        返回值：
            Balance (int): 账户网络原生代币余额。单位为 wei ，当出现异常时返回 None 。
        """

        try:
            Address = Web3.toChecksumAddress(Address)
            Balance = self.Eth.get_balance(Address)
            logger.success(
                f"\n[Chain][GetBalance]\n[Address]{Address}\n[Balance][{Balance} Wei]<=>[{Web3.fromWei(Balance,'ether')} Ether]\n{'-'*80}"
            )
            return Balance
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(f"\n[Chain][GetBalance]Failed\n[Address]{Address}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}")
            return None

    def GetCode(self, Address: str) -> str:
        """
        根据合约地址获取其已部署字节码。

        参数：
            Address (str): 合约地址

        返回值：
            Code (str): 合约已部署字节码。含 0x 前缀的十六进制形式，当出现异常时返回 None 。
        """

        try:
            Address = Web3.toChecksumAddress(Address)
            Code = self.Eth.get_code(Address).hex()
            logger.success(f"\n[Chain][GetCode]\n[Address]{Address}\n[Code]{Code}\n{'-'*80}")
            return Code
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(f"\n[Chain][GetCode]Failed\n[Address]{Address}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}")
            return None

    def GetStorage(self, Address: str, SlotIndex: int) -> str:
        """
        根据合约地址和存储插槽索引获取存储值。

        参数：
            Address (str): 合约地址
            SlotIndex (int): 存储插槽索引

        返回值：
            Data (str): 存储值。含 0x 前缀的十六进制形式，当出现异常时返回 None 。
        """

        try:
            Address = Web3.toChecksumAddress(Address)
            Data = self.Eth.get_storage_at(Address, SlotIndex).hex()
            logger.success(
                f"\n[Chain][GetStorage]\n[Address]{Address}\n[SlotIndex]{SlotIndex}\n[Value][Hex][{Data}]<=>[Dec][{int(Data,16)}]\n{'-'*80}"
            )
            return Data
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Chain][GetStorage]Failed\n[Address]{Address}\n[SlotIndex]{SlotIndex}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def DumpStorage(self, Address: str, Count: int) -> list:
        """
        根据合约地址和指定插槽数量值，从插槽 0 开始批量遍历存储插槽并获取值。

        参数：
            Address (str): 合约地址
            Count (int): 指定插槽数量值

        返回值：
            Data (List[str]): 存储值列表。含 0x 前缀的十六进制形式，当出现异常时返回 None 。
        """

        try:
            Address = Web3.toChecksumAddress(Address)
            Data = [self.Eth.get_storage_at(Address, i).hex() for i in range(Count)]
            Temp = '\n'.join([f"[Slot {i}]{Data[i]}" for i in range(len(Data))])
            logger.success(f"\n[Chain][DumpStorage]\n[Address]{Address}\n{Temp}\n{'-'*80}")
            return Data
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Chain][DumpStorage]Failed\n[Address]{Address}\n[slot 0 ... {Count-1}]\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def GetPublicKeyByTransactionHash(self, TransactionHash: str) -> tuple:
        """
        通过一笔已在链上确认的交易哈希，获取账户公钥。

        参数：
            TransactionHash (str): 交易哈希

        返回值：
            (Address, PublicKey) (tuple): 由账户地址和账户公钥组成的元组。当出现异常时返回 None 。
        """

        try:
            from eth_account._utils.signing import to_standard_v, extract_chain_id, serializable_unsigned_transaction_from_dict
            Transaction = self.Eth.get_transaction(TransactionHash)
            Signature = self.Eth.account._keys.Signature(vrs=(to_standard_v(extract_chain_id(Transaction.v)[1]), Web3.toInt(Transaction.r), Web3.toInt(Transaction.s)))
            UnsignedTransactionDict = {i: Transaction[i] for i in ['chainId', 'nonce', 'gasPrice' if int(
                Transaction.type, 0) != 2 else '', 'gas', 'to', 'value', 'accessList', 'maxFeePerGas', 'maxPriorityFeePerGas'] if i in Transaction}
            UnsignedTransactionDict['data'] = Transaction['input']
            UnsignedTransaction = serializable_unsigned_transaction_from_dict(UnsignedTransactionDict)
            Temp = Signature.recover_public_key_from_msg_hash(UnsignedTransaction.hash())
            PublicKey = str(Temp).replace('0x', '0x04')  # 比特币未压缩公钥格式
            Address = Temp.to_checksum_address()
            logger.success(
                f"\n[Chain][GetPublicKeyByTransactionHash]\n[TransactionHash]{TransactionHash}\n[Address]{Address}\n[PublicKey]{PublicKey}\n{'-'*80}"
            )
            return (Address, PublicKey)
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Chain][GetPublicKeyByTransactionHash]Failed\n[TransactionHash]{TransactionHash}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None


class Account():
    """
    Account 是账户实例，后续的交易将经由该指定账户发送至链上。
    """

    def __init__(self, Chain: Chain, PrivateKey: str):
        """
        初始化。通过私钥导入账户并与 Chain 实例绑定，后续的交易将经由该指定账户发送至链上。当导入账户失败时将会抛出异常。

        参数：
            Chain (Poseidon.Blockchain.Chain): 区块链实例
            PrivateKey (str): 账户私钥。不含 0x 前缀的十六进制形式。

        成员变量：
            EthAccount (eth_account.Account): eth_account 的原生 Account 对象实例
        """

        try:
            self.EthAccount = EthAccount.from_key(PrivateKey)
            self._Chain = Chain
            self._Eth = Chain.Eth
            self._Eth.default_account = self.EthAccount.address
            logger.success(f"\n[Account][Initialize]Successfully import account [{self.EthAccount.address}]\n{'-'*80}")
            self.GetSelfBalance()
        except:
            ExceptionInformation = format_exc()
            logger.error(f"\n[Account][Initialize]Failed to import account\n[ExceptionInformation]{ExceptionInformation}{'-'*80}")
            raise Exception("Failed to import account.")

    def GetSelfBalance(self) -> int:
        """
        获取自身账户的网络原生代币余额。

        返回值：
            Balance (int): 自身账户网络原生代币余额。单位为 wei ，当出现异常时返回 None 。
        """

        Balance = self._Chain.GetBalance(self.EthAccount.address)
        if Balance == 0:
            logger.warning(f"\n[Account][GetSelfBalance]\n[Warning]This account's balance is insufficient to send transactions\n{'-'*80}")
        return Balance

    def Transfer(self, To: str, Value: int, GasLimit: int = 100000, Data: str = "0x") -> dict:
        """
        向指定账户转账指定数量的网络原生代币，可附带信息。若 90 秒内交易未确认则作超时处理。

        参数：
            To (str): 接收方地址
            Value (int): 发送的网络原生代币数量。单位为 wei 。
            GasLimit (可选)(int): Gas 最大使用量。单位为 wei ，默认为 100000 wei 。
            Data (可选)(str): 交易数据。含 0x 前缀的十六进制形式，默认值为 "0x" 。

        返回值：
            TransactionInformation (dict): 交易信息构成的字典，通过 Chain.GetTransactionInformationByHash 获取。当出现异常时返回 None 。
        """

        try:
            From = self.EthAccount.address
            To = Web3.toChecksumAddress(To)
            Txn = {
                "chainId": self._Chain.ChainId,
                "from": From,
                "to": Web3.toChecksumAddress(To),
                "value": Value,
                "gas": GasLimit,
                "gasPrice": self._Eth.gas_price,
                "nonce": self._Eth.get_transaction_count(From),
                "data": Data,
            }
            SignedTxn = self.EthAccount.sign_transaction(Txn)
            TransactionHash = self._Eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            Txn["gasPrice"] = f'{Web3.fromWei(Txn["gasPrice"],"gwei")} Gwei'
            logger.info(f"\n[Account][Transfer]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}\n{'-'*80}")
            TransactionInformation = self._Chain.GetTransactionInformationByHash(TransactionHash)
            return TransactionInformation
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Account][Transfer]Failed\n[From]{From}\n[To]{To}\n[Value]{Value}[GasLimit]{GasLimit}\n[Data]{Data}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def SendTransaction(self, To: str, Data: str, Value: int = 0, GasLimit: int = 1000000) -> dict:
        """
        以传统方式发送一笔自定义交易。若 90 秒内交易未确认则作超时处理。

        参数：
            To (str): 接收方地址
            Data (str): 交易数据。含 0x 前缀的十六进制形式。
            Value (可选)(int): 随交易发送的网络原生代币数量。单位为 wei ，默认为 0 wei 。
            GasLimit (可选)(int): Gas 最大使用量。单位为 wei ，默认为 1000000 wei 。

        返回值：
            TransactionInformation (dict): 交易信息构成的字典，通过 Chain.GetTransactionInformationByHash 获取。当出现异常时返回 None 。
        """

        try:
            From = self.EthAccount.address
            To = Web3.toChecksumAddress(To)
            Txn = {
                "chainId": self._Chain.ChainId,
                "from": From,
                "to": To,
                "value": Value,
                "gas": GasLimit,
                "gasPrice": self._Eth.gas_price,
                "nonce": self._Eth.get_transaction_count(From),
                "data": Data,
            }
            SignedTxn = self.EthAccount.sign_transaction(Txn)
            TransactionHash = self._Eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            Txn["gasPrice"] = f'{Web3.fromWei(Txn["gasPrice"],"gwei")} Gwei'
            logger.info(f"\n[Account][SendTransaction][Traditional]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}\n{'-'*80}")
            TransactionInformation = self._Chain.GetTransactionInformationByHash(TransactionHash)
            return TransactionInformation
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Account][SendTransaction][Traditional]Failed\n[From]{From}\n[To]{To}\n[Value]{Value}[GasLimit]{GasLimit}\n[Data]{Data}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def SendTransactionByEIP1559(self, To: str, Data: str, Value: int = 0, GasLimit: int = 1000000) -> dict:
        """
        以 EIP-1559 方式发送一笔自定义交易。若 90 秒内交易未确认则作超时处理。

        参数：
            To (str): 接收方地址
            Data (str): 交易数据。含 0x 前缀的十六进制形式。
            Value (可选)(int): 随交易发送的网络原生代币数量。单位为 wei ，默认为 0 wei 。
            GasLimit (可选)(int): Gas 最大使用量。单位为 wei ，默认为 1000000 wei 。

        返回值：
            TransactionInformation (dict): 交易信息构成的字典，通过 Chain.GetTransactionInformationByHash 获取。当出现异常时返回 None 。
        """

        try:
            From = self.EthAccount.address
            To = Web3.toChecksumAddress(To)
            BaseFee = self._Eth.gas_price
            MaxPriorityFee = self._Eth.max_priority_fee + Web3.toWei(1, "gwei")
            Txn = {
                "chainId": self._Chain.ChainId,
                "from": From,
                "to": To,
                "value": Value,
                "gas": GasLimit,
                "maxFeePerGas": BaseFee + MaxPriorityFee,
                "maxPriorityFeePerGas": MaxPriorityFee,
                "nonce": self._Eth.get_transaction_count(From),
                "data": Data
            }
            SignedTxn = self.EthAccount.sign_transaction(Txn)
            TransactionHash = self._Eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            Txn["maxFeePerGas"] = f'{Web3.fromWei(Txn["maxFeePerGas"],"gwei")} Gwei'
            Txn["maxPriorityFeePerGas"] = f'{Web3.fromWei(Txn["maxPriorityFeePerGas"],"gwei")} Gwei'
            logger.info(f"\n[Account][SendTransaction][EIP-1559]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}\n{'-'*80}")
            TransactionInformation = self._Chain.GetTransactionInformationByHash(TransactionHash)
            return TransactionInformation
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Account][SendTransaction][EIP-1559]Failed\n[From]{From}\n[To]{To}\n[Value]{Value}[GasLimit]{GasLimit}\n[Data]{Data}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def DeployContract(self, ABI: dict, Bytecode: str, Value: int = 0, *Arguments: Optional[Any]) -> dict:
        """
        部署合约。若 90 秒内交易未确认则作超时处理。

        参数：
            ABI (dict): 合约 ABI
            Bytecode (str): 合约部署字节码。含 0x 前缀的十六进制形式。
            Value (可选)(int): 随交易发送给合约的网络原生代币数量。单位为 wei ，默认为 0 wei 。
            *Arguments (Optional[Any]): 传给合约构造函数的参数，默认为空。

        返回值：
            TransactionInformation (dict): 交易信息构成的字典，通过 Chain.GetTransactionInformationByHash 获取。当出现异常时返回 None 。
            当合约部署成功时，字典中会额外添加"Contract"字段，该变量是已实例化的 Contract 对象，失败时为 None。
        """

        try:
            DeployingContract = self._Eth.contract(abi=ABI, bytecode=Bytecode)
            TransactionData = DeployingContract.constructor(*Arguments).buildTransaction({"gasPrice": self._Eth.gas_price, "value": Value})
            Txn = {
                "chainId": self._Chain.ChainId,
                "from": self.EthAccount.address,
                "value": TransactionData["value"],
                "gas": TransactionData["gas"],
                "gasPrice": TransactionData["gasPrice"],
                "nonce": self._Eth.get_transaction_count(self.EthAccount.address),
                "data": TransactionData["data"]
            }
            SignedTxn = self.EthAccount.sign_transaction(Txn)
            TransactionHash = self._Eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            Txn["gasPrice"] = f'{Web3.fromWei(Txn["gasPrice"],"gwei")} Gwei'
            logger.info(f"\n[Account][DeployContract]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}\n{'-'*80}")
            TransactionInformation = self._Chain.GetTransactionInformationByHash(TransactionHash)
            if TransactionInformation["Status"]:
                DeployedContract = Contract(self, TransactionInformation["ContractAddress"], ABI)
                TransactionInformation["Contract"] = DeployedContract
                return TransactionInformation
            else:
                TransactionInformation["Contract"] = None
                return TransactionInformation
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Account][DeployContract]Failed\n[Value]{Value}\n[ABI]{ABI}\n[Bytecode]{Bytecode}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def DeployContractWithoutABI(self, Bytecode: str, Value: int = 0, GasLimit: int = 10000000) -> dict:
        """
        在没有 ABI 的情况下，仅使用字节码来部署合约。若 90 秒内交易未确认则作超时处理。

        参数：
            Bytecode (str): 合约部署字节码。含 0x 前缀的十六进制形式。
            Value (可选)(int): 随交易发送给合约的网络原生代币数量。单位为 wei ，默认为 0 wei 。
            GasLimit (可选)(int): Gas 最大使用量。单位为 wei ，默认为 10000000 wei 。

        返回值：
            TransactionInformation (dict): 交易信息构成的字典，通过 Chain.GetTransactionInformationByHash 获取。当出现异常时返回 None 。
        """

        try:
            Txn = {
                "chainId": self._Chain.ChainId,
                "from": self.EthAccount.address,
                "value": Value,
                "gas": GasLimit,
                "gasPrice": self._Eth.gas_price,
                "nonce": self._Eth.get_transaction_count(self.EthAccount.address),
                "data": Bytecode,
            }
            SignedTxn = self.EthAccount.sign_transaction(Txn)
            TransactionHash = self._Eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            Txn["gasPrice"] = f'{Web3.fromWei(Txn["gasPrice"],"gwei")} Gwei'
            logger.info(f"\n[Account][DeployContractWithoutABI]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}\n{'-'*80}")
            TransactionInformation = self._Chain.GetTransactionInformationByHash(TransactionHash)
            return TransactionInformation
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Account][DeployContractWithoutABI]Failed\n[Value]{Value}\n[GasLimit]{GasLimit}\n[Bytecode]{Bytecode}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def SignMessage(self, Message: str) -> dict:
        """
        消息字符串进行签名。

        参数：
            Message (str): 待签名消息字符串

        返回值：
            SignatureData (str): 签名数据构成的字典。当出现异常时返回 None 。
            {"Address"|"Message"|"MessageHash"|"Signature"|"R"|"S"|"V"}
        """

        try:
            from eth_account.messages import encode_defunct
            SignedMessage = self.EthAccount.sign_message(encode_defunct(text=Message))
            MessageHash = SignedMessage.messageHash.hex()
            Signature = SignedMessage.signature.hex()
            R = hex(SignedMessage.r)
            S = hex(SignedMessage.s)
            V = hex(SignedMessage.v)
            logger.success(
                f"\n[Account][SignMessage]\n[Address]{self.EthAccount.address}\n[Message]{Message}\n[MessageHash]{MessageHash}\n[Signature]{Signature}\n[R]{R}\n[S]{S}\n[V]{V}\n{'-'*80}"
            )
            return {"Address": self.EthAccount.address, "Message": Message, "MessageHash": MessageHash, "Signature": Signature, "R": R, "S": S, "V": V}
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Account][SignMessage]Failed to sign message\n[Address]{self.EthAccount.address}\n[Message]{Message}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def SignMessageHash(self, MessageHash: str) -> dict:
        """
        对消息哈希进行签名。

        参数：
            MessageHash (str): 待签名消息哈希

        返回值：
            SignatureData (str): 签名数据构成的字典。当出现异常时返回 None 。
            {"Address"|"MessageHash"|"Signature"|"R"|"S"|"V"}
        """

        try:
            SignedMessage = self.EthAccount.signHash(MessageHash)
            Signature = SignedMessage.signature.hex()
            R = hex(SignedMessage.r)
            S = hex(SignedMessage.s)
            V = hex(SignedMessage.v)
            logger.success(
                f"\n[Account][SignMessageHash]\n[Address]{self.EthAccount.address}\n[MessageHash]{MessageHash}\n[Signature]{Signature}\n[R]{R}\n[S]{S}\n[V]{V}\n{'-'*80}"
            )
            return {"Address": self.EthAccount.address, "MessageHash": MessageHash, "Signature": Signature, "R": R, "S": S, "V": V}
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Account][SignMessageHash]Failed\n[Address]{self.EthAccount.address}\n[MessageHash]{MessageHash}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None


class Contract():
    """
    Contract 是合约实例，作为与指定合约进行交互的基础。
    """

    def __init__(self, Account: Account, Address: str, ABI: dict):
        """
        初始化。通过合约地址与 ABI 来实例化合约，并与 Account 绑定，后续所有对该合约的调用都会由这一账户发起。当合约实例化失败时会抛出异常。

        参数：
            Account (Poseidon.Blockchain.Account): 账户实例
            Address (str): 合约地址
            ABI (str): 合约 ABI

        成员变量：
            Instance (Web3.eth.Contract): web3.py 原生 contract 对象实例
            Address (str): 合约地址
        """

        try:
            self._Account = Account
            self._Eth = Account._Eth
            self.Address = Web3.toChecksumAddress(Address)
            self.Instance = self._Eth.contract(address=self.Address, abi=ABI)
            logger.success(f"\n[Contract][Initialize]Successfully instantiated contract [{self.Address}]\n{'-'*80}")
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Contract][Initialize]Failed to instantiated contract [{self.Address}]\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            raise Exception("Failed to instantiate contract.")

    def CallFunction(self, FunctionName: str, *FunctionArguments: Optional[Any]) -> dict:
        """
        通过传入函数名及参数来调用该合约内的函数。

        参数：
            FunctionName (str): 函数名称
            *FunctionArguments (Optional[Any]): 函数参数，默认为空。

        返回值：
            TransactionInformation (dict): 交易信息构成的字典，通过 Chain.GetTransactionInformationByHash 获取。当出现异常时返回 None 。
        """

        TransactionData = self.Instance.functions[FunctionName](*FunctionArguments).buildTransaction({"gasPrice": self._Eth.gas_price})
        logger.info(f"\n[Contract][CallFunction]\n[ContractAddress]{self.Address}\n[Function]{FunctionName}{FunctionArguments}\n{'-'*80}")
        TransactionInformation = self._Account.SendTransaction(self.Address, TransactionData["data"], TransactionData["value"], TransactionData["gas"])
        return TransactionInformation

    def CallFunctionWithValueAndGasLimit(self, Value: int, GasLimit: int, FunctionName: str, *FunctionArguments: Optional[Any]) -> dict:
        """
        通过传入函数名及参数来调用该合约内的函数。支持自定义 Value 和 GasLimit 。

        参数：
            Value (int): 随交易发送的网络原生代币数量。单位为 wei 。
            GasLimit (int): Gas 最大使用量。单位为 wei 。
            FunctionName (str): 函数名称
            *FunctionArguments (Optional[Any]): 函数参数，默认为空。

        返回值：
            TransactionInformation (dict): 交易信息构成的字典，通过 Chain.GetTransactionInformationByHash 获取。当出现异常时返回 None 。
        """

        TransactionData = self.Instance.functions[FunctionName](*FunctionArguments).buildTransaction({"gasPrice": self._Eth.gas_price, "gas": GasLimit, "value": Value})
        logger.info(
            f"\n[Contract][CallFunctionWithValueAndGasLimit]\n[ContractAddress]{self.Address}\n[Function]{FunctionName}{FunctionArguments}\n[Value]{TransactionData['value']} [GasLimit]{TransactionData['gas']}\n{'-'*80}"
        )
        TransactionInformation = self._Account.SendTransaction(self.Address, TransactionData["data"], TransactionData["value"], TransactionData["gas"])
        return TransactionInformation

    def ReadOnlyCallFunction(self, FunctionName: str, *FunctionArguments: Optional[Any]) -> Any:
        """
        通过传入函数名及参数来调用该合约内的只读函数。

        参数：
            FunctionName (str): 函数名称
            *FunctionArguments (Optional[Any]): 函数参数，默认为空。

        返回值：
            Result (Any): 调用函数后得到的返回值。当出现异常时返回 None 。
        """

        try:
            Result = self.Instance.functions[FunctionName](*FunctionArguments).call()
            logger.success(
                f"\n[Contract][ReadOnlyCallFunction]\n[ContractAddress]{self.Address}\n[Function]{FunctionName}{FunctionArguments}\n[Result]{Result}\n{'-'*80}"
            )
            return Result
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Contract][ReadOnlyCallFunction]Failed\n[ContractAddress]{self.Address}\n[Function]{FunctionName}{FunctionArguments}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    def EncodeABI(self, FunctionName: str, *FunctionArguments: Optional[Any]) -> str:
        """
        通过传入函数名及参数进行编码，相当于生成调用该函数的 CallData 。

        参数：
            FunctionName (str): 函数名称
            *FunctionArguments (Optional[Any]): 函数参数，默认为空。

        返回值：
            CallData (str): 调用数据编码。含 0x 前缀的十六进制形式。当出现异常时返回 None 。
        """

        try:
            CallData = self.Instance.encodeABI(fn_name=FunctionName, args=FunctionArguments)
            logger.success(
                f"\n[Contract][EncodeABI]\n[ContractAddress]{self.Address}\n[Function]{FunctionName}{FunctionArguments}\n[CallData]{CallData}\n{'-'*80}"
            )
            return CallData
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[Contract][EncodeABI]Failed\n[ContractAddress]{self.Address}\n[Function]{FunctionName}{FunctionArguments}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None


class BlockchainUtils():
    """
    通用工具集，整合了常用的链下操作。静态类，无需实例化。
    """

    @staticmethod
    def SwitchSolidityVersion(SolidityVersion: str):
        """
        设置当前使用的 Solidity 版本，若该版本未安装则会自动安装。

        参数：
            SolidityVersion (str): Solidity 版本号
        """

        try:
            from solcx import install_solc, set_solc_version
            install_solc(SolidityVersion)
            set_solc_version(SolidityVersion)
            logger.success(f"\n[BlockchainUtils][SwitchSolidityVersion]Current Solidity Version [{SolidityVersion}]\n{'-'*80}")
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][SwitchSolidityVersion]Failed to switch to version [{SolidityVersion}]\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )

    @staticmethod
    def Compile(FileCourse: str, ContractName: str, SolidityVersion: Optional[str] = None, AllowPaths: Optional[str] = None, Optimize: Optional[bool] = False) -> tuple:
        """
        根据给定的参数使用 py-solc-x 编译合约。当编译失败时会抛出异常。

        参数：
            FileCourse (str): 合约文件完整路径。当合约文件与脚本文件在同一目录下时可直接使用文件名。
            ContractName (str): 要编译的合约名称
            SolidityVersion (Optional[str]): 指定使用的 Solidity 版本。若不指定则会使用当前已激活的 Solidity 版本进行编译。默认为 None 。
            AllowPaths (Optional[str]): 指定许可路径。在编译时可能会出现 AllowPaths 相关错误可在这里解决。默认为 None 。
            Optimize (Optional[bool]): 是否开启优化器。默认为 False 。

        返回值：
            (ABI, Bytecode) (tuple): 由 ABI 和 Bytecode 组成的元组
        """

        try:
            from solcx import compile_source
            with open(FileCourse, "r", encoding="utf-8") as sol:
                CompiledSol = compile_source(sol.read(), solc_version=SolidityVersion, allow_paths=AllowPaths, optimize=Optimize, output_values=['abi', 'bin'])
            ContractData = CompiledSol[f'<stdin>:{ContractName}']
            ABI = ContractData['abi']
            Bytecode = ContractData['bin']
            with open(f'{ContractName}_ABI.json', 'w', encoding="utf-8") as f:
                dump(ABI, f, indent=4)
            logger.success(
                f"\n[BlockchainUtils][Compile]\n[FileCourse]{FileCourse}\n[ContractName]{ContractName}\n[ABI]{ABI}\n[Bytecode]{Bytecode}\n{'-'*80}"
            )
            return (ABI, Bytecode)
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][Compile]Failed\n[FileCourse]{FileCourse}\n[ContractName]{ContractName}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            raise Exception("Failed to compile the contract.")

    @staticmethod
    def CreateNewAccount() -> tuple:
        """
        创建新账户。

        返回值：
            (Address, PrivateKey) (tuple): 由账户地址和私钥组成的元组
        """

        Temp = EthAccount.create()
        Address = Web3.toChecksumAddress(Temp.address)
        PrivateKey = Temp.privateKey.hex()
        logger.success(f"\n[BlockchainUtils][CreateNewAccount]\n[Address]{Address}\n[PrivateKey]{PrivateKey}\n{'-'*80}")
        return (Address, PrivateKey)

    @staticmethod
    def MnemonicToAddressAndPrivateKey(Mnemonic: str) -> tuple:
        """
        将助记词转换为账户地址与私钥。参考 BIP-39 标准。

        参数：
            Mnemonic (str): 助记词字符串。以空格进行分隔。

        返回值：
            (Address, PrivateKey) (tuple): 由账户地址和私钥组成的元组。当出现异常时返回 None 。
        """

        try:
            EthAccount.enable_unaudited_hdwallet_features()
            Temp = EthAccount.from_mnemonic(Mnemonic)
            Address = Web3.toChecksumAddress(Temp.address)
            PrivateKey = Temp.privateKey.hex()
            logger.success(
                f"\n[BlockchainUtils][MnemonicToAddressAndPrivateKey]\n[Mnemonic]{Mnemonic}\n[Address]{Address}\n[PrivateKey]{PrivateKey}\n{'-'*80}"
            )
            return (Address, PrivateKey)
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][MnemonicToAddressAndPrivateKey]Failed\n[Mnemonic]{Mnemonic}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    @staticmethod
    def RecoverMessage(Message: str, Signature: str) -> str:
        """
        通过消息原文和签名还原出签署者的账户地址。

        参数：
            Message (str): 消息原文
            Signature (str): 签名

        返回值：
            Signer (str): 签署者的账户地址。当出现异常时返回 None 。
        """

        try:
            from eth_account.messages import encode_defunct
            Signer = EthAccount.recover_message(encode_defunct(text=Message), signature=Signature)
            logger.success(f"\n[BlockchainUtils][RecoverMessage]\n[Message]{Message}\n[Signature]{Signature}\n[Signer]{Signer}\n{'-'*80}")
            return Signer
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][RecoverMessage]Failed\n[Message]{Message}\n[Signature]{Signature}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    @staticmethod
    def RecoverMessageHash(MessageHash: str, Signature: str) -> str:
        """
        通过消息哈希和签名还原出签署者的账户地址。

        参数：
            MessageHash (str): 消息哈希
            Signature (str): 签名

        返回值：
            Signer (str): 签署者的账户地址。当出现异常时返回 None 。
        """

        try:
            Signer = EthAccount.recoverHash(MessageHash, signature=Signature)
            logger.success(
                f"\n[BlockchainUtils][RecoverMessageByHash]\n[MessageHash]{MessageHash}\n[Signature]{Signature}\n[Signer]{Signer}\n{'-'*80}"
            )
            return Signer
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][RecoverMessageByHash]Failed\n[MessageHash]{MessageHash}\n[Signature]{Signature}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    @staticmethod
    def RecoverRawTransaction(RawTransactionData: str) -> str:
        """
        获取签署此交易的账户地址。

        参数：
            RawTransactionData (str): 原生交易数据。含 0x 前缀的十六进制形式。

        返回值：
            Address (str): 账户地址。当出现异常时返回 None 。
        """

        try:
            Address = EthAccount.recover_transaction(RawTransactionData)
            logger.success(f"\n[BlockchainUtils][RecoverRawTransaction]\n[RawTransactionData]{RawTransactionData}\n[Address]{Address}\n{'-'*80}")
            return Address
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][RecoverRawTransaction]Failed\n[RawTransactionData]{RawTransactionData}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    @staticmethod
    def CrackSelector(SourceFunctionName: str, SourceFunctionParameters: List[str], ToGenerateFunctionParameters: List[str]) -> str:
        """
        根据源函数名、参数与想要碰撞生成的函数的参数，碰撞生成出一个函数名，以使得这两个函数的选择器签名相等。

        参数：
            SourceFunctionName (str): 目标函数名
            SourceFunctionParameters (List[str]): 目标函数参数列表
            ToGenerateFunctionParameters (List[str]): 想要碰撞生成的函数的参数列表

        返回值：
            ToGenerateFunction (str): 碰撞出的函数的名称与参数完整表示。当出现异常时返回 None 。
        """
        from pwn import pwnlib
        from pwnlib.util.iters import mbruteforce

        def Crack(SourceFunctionSelector: str, Temp: str) -> str:
            Charset = "0123456789abcdef"
            X = mbruteforce(lambda x: Web3.keccak(f"function_{x}({Temp})".encode())[:4].hex() == SourceFunctionSelector, Charset, 8, method='fixed')
            return f"function_{X}({Temp})"

        try:
            SourceFunctionSelector = Web3.keccak(f"{SourceFunctionName}({','.join(SourceFunctionParameters)})".encode())[:4].hex()
            Temp = ','.join(ToGenerateFunctionParameters)
            logger.info(
                f"\n[BlockchainUtils][CrackSelector]\n[SourceFunction]{SourceFunctionName}({','.join(SourceFunctionParameters)})\n[SourceFunctionSelector]{SourceFunctionSelector}\n[ToGenerateFunction]function_?({Temp})\nCrack start..."
            )
            ToGenerateFunction = Crack(SourceFunctionSelector, Temp)
            ToGenerateFunctionSelector = Web3.keccak(ToGenerateFunction.encode())[:4].hex()
            logger.success(
                f"\n[BlockchainUtils][CrackSelector]\n[ToGenerateFunction]{ToGenerateFunction}\n[ToGenerateFunctionSelector]{ToGenerateFunctionSelector}\n{'-'*80}"
            )
            return ToGenerateFunction
        except Exception:
            ExceptionInformation = format_exc()
            Temp = ','.join(ToGenerateFunctionParameters)
            logger.error(
                f"\n[BlockchainUtils][CrackSelector]Failed\n[SourceFunction]{SourceFunctionName}({','.join(SourceFunctionParameters)})\n[ToGenerateFunction]{f'function_?({Temp})'}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    @staticmethod
    def AssemblyToBytecode(Assembly: str) -> str:
        """
        将 EVM Assembly 转为 EVM Bytecode 。

        参数：
            Assembly (str): EVM Assembly

        返回值：
            Bytecode (str): EVM Bytecode 。含 0x 前缀的六进制形式。当出现异常时返回 None 。
        """

        try:
            from pyevmasm import assemble_hex
            Bytecode = assemble_hex(Assembly)
            logger.success(f"\n[BlockchainUtils][AssemblyToBytecode]\n[Bytecode]{Bytecode}\n{'-'*80}")
            return Bytecode
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][AssemblyToBytecod]Failed\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    @staticmethod
    def BytecodeToAssembly(Bytecode: str) -> str:
        """
        将 EVM Bytecode 转为 EVM Assembly 。

        参数：
            Bytecode (str): EVM Bytecode 。含 0x 前缀的十六进制形式。

        返回值：
            Assembly (str): EVM Assembly 。当出现异常时返回 None 。
        """

        try:
            from pyevmasm import disassemble_hex
            Assembly = disassemble_hex(Bytecode)
            logger.success(f"\n[BlockchainUtils][AssemblyToBytecode]\n[Assembly]\n{Assembly}\n{'-'*80}")
            return Assembly
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][BytecodeToAssembly]Failed\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    @staticmethod
    def SignatureToRSV(Signature: str) -> dict:
        """
        将签名解析成 R S V 。

        参数：
            Signature (str): 签名。含 0x 前缀的十六进制形式。

        返回值：
            Result (dict): 解析结果。当出现异常时返回 None 。
            {"Signature"|"R"|"S"|"V"}
        """

        try:
            Signature = hex(int(Signature, 16))
            assert (len(Signature) == 130 + 2)
            R = '0x' + Signature[2:66]
            S = '0x' + Signature[66:-2]
            V = '0x' + Signature[-2:]
            logger.success(f"\n[BlockchainUtils][SignatureToRSV]\n[Signature]{Signature}\n[R]{R}\n[S]{S}\n[V]{V}\n{'-'*80}")
            Result = {"Signature": Signature, "R": R, "S": S, "V": V}
            return Result
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][SignatureToRSV]Failed\n[Signature]{Signature}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None

    @staticmethod
    def RSVToSignature(R: str, S: str, V: str) -> dict:
        """
        将 R S V 合并成签名。

        参数：
            R (str): 签名 r 值。含 0x 前缀的十六进制形式。
            S (str): 签名 s 值。含 0x 前缀的十六进制形式。
            V (str): 签名 v 值。含 0x 前缀的十六进制形式。

        返回值：
            Result (dict): 合并结果。当出现异常时返回 None 。
            {"R"|"S"|"V"|"Signature"}
        """

        try:
            R = hex(int(R, 16))
            S = hex(int(S, 16))
            V = hex(int(V, 16))
            assert (len(R) == 64 + 2 and len(S) == 64 + 2 and len(V) == 2 + 2)
            Signature = '0x' + R[2:] + S[2:] + V[2:]
            logger.success(f"\n[BlockchainUtils][RSVToSignature]\n[R]{R}\n[S]{S}\n[V]{V}\n[Signature]{Signature}\n{'-'*80}")
            Result = {"R": R, "S": S, "V": V, "Signature": Signature}
            return Result
        except Exception:
            ExceptionInformation = format_exc()
            logger.error(
                f"\n[BlockchainUtils][RSVToSignature]Failed\n[R]{R}\n[S]{S}\n[V]{V}\n[ExceptionInformation]{ExceptionInformation}{'-'*80}"
            )
            return None



"""
本模块用于解决常见的密码学问题。
"""

import gmpy2
import base64
from Crypto.Cipher import AES as aes
from Crypto.Util.number import long_to_bytes, bytes_to_long
from typing import Optional, List


class MiscUtils():
    """
    本模块用于处理进制转换和常用哈希。静态类，无需实例化。
    """

    @staticmethod
    def Binary_String(Binary: str) -> str:
        """
        用于将形如"1010...0101"的二进制字符串按照"8位1字符"的规则转换为字符串。

        参数：
            Binary (str): 二进制字符串

        返回值：
            String (str): 转换得到的字符串
        """

        if len(Binary) % 8 != 0:
            raise Exception("二进制字符串的长度应该为8的倍数")
        String = "".join([chr(int(Binary[i:i + 8], 2)) for i in range(0, len(Binary), 8)])
        return String

    @staticmethod
    def Binary_Dec(Binary: str) -> int:
        """
        用于将形如"1010...0101"的二进制字符串转换为十进制整数形式。

        参数：
            Binary (str): 二进制字符串

        返回值：
            Dec (int): 转换得到的十进制整数
        """

        Dec = int(Binary, 2)
        return Dec

    @staticmethod
    def Binary_Hex(Binary: str) -> str:
        """
        用于将形如"1010...0101"的二进制字符串转换为十六进制字符串形式（含 0x 前缀）。

        参数：
            Binary (str): 二进制字符串

        返回值：
            Hex (str): 转换得到的十六进制字符串
        """

        Hex = hex(int(Binary, 2))
        return Hex

    @staticmethod
    def Dec_String(Dec: int) -> str:
        """
        用于将十进制整数转换为字符串（UTF-8 字符集）。

        参数：
            Dec (int): 十进制整数

        返回值：
            String (str): 转换得到的字符串
        """

        from Crypto.Util.number import long_to_bytes
        String = long_to_bytes(Dec).decode()
        return String

    @staticmethod
    def Dec_Binary(Dec: int) -> str:
        """
        用于将十进制整数转换为二进制字符串形式（含 0b 前缀）。

        参数：
            Dec (int): 十进制整数

        返回值：
            Binary (str): 转换得到的二进制字符串
        """

        Binary = bin(Dec)
        return Binary

    @staticmethod
    def Dec_Hex(Dec: int) -> str:
        """
        用于将十进制整数转换为十六进制字符串形式（含 0x 前缀）。

        参数：
            Dec (int): 十进制整数

        返回值：
            Hex (str): 转换得到的十六进制字符串
        """

        Hex = hex(Dec)
        return Hex

    @staticmethod
    def Hex_String(Hex: str) -> str:
        """
        用于将形如"0a0b0c...1c1b1a"的十六进制字符串按照"2位1字符"的规则转换为字符串。

        参数：
            Hex (str): 十六进制字符串

        返回值：
            String (str): 转换得到的字符串
        """

        if len(Hex) % 2 != 0:
            raise Exception("十六进制字符串的长度应该为2的倍数")
        String = "".join([chr(int(Hex[i:i + 2], 16)) for i in range(0, len(Hex), 2)])
        return String

    @staticmethod
    def Hex_Binary(Hex: str) -> str:
        """
        用于将形如"0a0b0c...1c1b1a"的十六进制字符串为二进制字符串形式（含 0b 前缀）。

        参数：
            Hex (str): 十六进制字符串

        返回值：
            Binary (str): 转换得到的二进制字符串
        """

        Binary = bin(int(Hex, 16))
        return Binary

    @staticmethod
    def Hex_Dec(Hex: str) -> int:
        """
        用于将形如"0a0b0c...1c1b1a"的十六进制字符串为十进制整数形式。

        参数：
            Hex (str): 十六进制字符串

        返回值：
            Dec (int): 转换得到的十进制整数
        """

        Dec = int(Hex, 16)
        return Dec

    @staticmethod
    def SHA1(Text: str) -> str:
        """
        用于获取字符串的 SHA1 哈希值。

        参数：
            Text (str): 字符串

        返回值：
            Hash (str): 该字符串的 SHA1 哈希值（十六进制字符串，不含 0x 前缀）
        """

        from hashlib import sha1
        Hash = sha1(Text.encode()).hexdigest()
        return Hash

    @staticmethod
    def SHA256(Text: str) -> str:
        """
        用于获取字符串的 SHA256 哈希值。

        参数：
            Text (str): 字符串

        返回值：
            Hash (str): 该字符串的 SHA256 哈希值（十六进制字符串，不含 0x 前缀）
        """

        from hashlib import sha256
        Hash = sha256(Text.encode()).hexdigest()
        return Hash

    @staticmethod
    def SHA512(Text: str) -> str:
        """
        用于获取字符串的 SHA512 哈希值。

        参数：
            Text (str): 字符串

        返回值：
            Hash (str): 该字符串的 SHA512 哈希值（十六进制字符串，不含 0x 前缀）
        """

        from hashlib import sha512
        Hash = sha512(Text.encode()).hexdigest()
        return Hash

    @staticmethod
    def MD5(Text: str) -> str:
        """
        用于获取字符串的 MD5 哈希值。

        参数：
            Text (str): 字符串

        返回值：
            Hash (str): 该字符串的 MD5 哈希值（十六进制字符串，不含 0x 前缀）
        """

        from hashlib import md5
        Hash = md5(Text.encode()).hexdigest()
        return Hash


class ModernCryptoUtils():
    """
    本模块用于解决现代密码学问题。
    """

    @staticmethod
    def Base64_Encrypt(Text: str) -> str:
        """
        用于对字符串进行 Base64 编码。

        参数：
            Text (str): 待编码的字符串

        返回值：
            EncryptedText (str): Base64 编码后的字符串
        """

        EncryptedText = base64.b64encode(Text.encode()).decode()
        return EncryptedText

    @staticmethod
    def Base64_Decrypt(Text: str) -> str:
        """
        用于对 Base64 编码的字符串进行解码。

        参数：
            Text (str): 待解码的 Base64 编码字符串

        返回值：
            DecryptedText (str): Base64 解码后的字符串
        """

        DecryptedText = base64.b64decode(Text.encode()).decode()
        return DecryptedText

    @staticmethod
    def Base32_Encrypt(Text: str) -> str:
        """
        用于对字符串进行 Base32 编码。

        参数：
            Text (str): 待编码的字符串

        返回值：
            EncryptedText (str): Base32 编码后的字符串
        """

        EncryptedText = base64.b32encode(Text.upper().encode()).decode()
        return EncryptedText

    @staticmethod
    def Base32_Decrypt(Text: str) -> str:
        """
        用于对 Base32 编码的字符串进行解码。

        参数：
            Text (str): 待解码的 Base32 编码字符串

        返回值：
            DecryptedText (str): Base32 解码后的字符串
        """

        DecryptedText = base64.b32decode(Text.upper().encode()).decode()
        return DecryptedText

    @staticmethod
    def Base16_Encrypt(Text: str) -> str:
        """
        用于对字符串进行 Base16 编码。

        参数：
            Text (str): 待编码的字符串

        返回值：
            EncryptedText (str): Base16 编码后的字符串
        """

        EncryptedText = base64.b16encode(Text.upper().encode()).decode()
        return EncryptedText

    @staticmethod
    def Base16_Decrypt(Text: str) -> str:
        """
        用于对 Base16 编码的字符串进行解码。

        参数：
            Text (str): 待解码的 Base16 编码字符串

        返回值：
            DecryptedText (str): Base16 解码后的字符串
        """

        DecryptedText = base64.b16decode(Text.upper().encode()).decode()
        return DecryptedText

    @staticmethod
    def AES_Padding(Text: str, BlockSize: int = 16) -> bytes:
        """
        用于对字符串进行 zeropadding 处理。

        参数：
            Text (str): 待 padding 的字符串
            BlockSize (可选)(int): 块大小（单位为字节），默认为16字节

        返回值：
            Fill (bytes): padding 后的字节数据
        """

        Fill = Text.encode()
        while len(Fill) % BlockSize != 0:
            Fill += b'\x00'
        return Fill

    @staticmethod
    def AES_Encrypt(Text: str, Key: str, BlockSize: int = 16) -> str:
        """
        用于对字符串进行 AES 加密（仅支持 ECB zeropadding 模式）。

        参数：
            Text (str): 待进行 AES 加密的字符串
            Key (str): 加密密钥
            BlockSize (可选)(int): 块大小（单位为字节），默认为16字节

        返回值：
            EncryptedText (str): AES 加密后的密文（Base64 编码形式）
        """

        AES = aes.new(ModernCryptoUtils.AES_Padding(Key, BlockSize), aes.MODE_ECB)
        EncryptedText = base64.b64encode(AES.encrypt(ModernCryptoUtils.AES_Padding(Text, BlockSize))).decode()
        return EncryptedText

    @staticmethod
    def AES_Decrypt(Text: str, Key: str, BlockSize: int = 16) -> str:
        """
        用于对 AES 密文进行解密（仅支持 ECB zeropadding 模式）。

        参数：
            Text (str): 待解密的 AES 密文（Base64 编码形式）
            Key (str): 解密密钥
            BlockSize (可选)(int): 块大小（单位为字节），默认为16字节

        返回值：
            DecryptedText (str): AES 解密后得到的原文
        """

        AES = aes.new(ModernCryptoUtils.AES_Padding(Key, BlockSize), aes.MODE_ECB)
        DecryptedText = AES.decrypt(base64.b64decode(ModernCryptoUtils.AES_Padding(Text, BlockSize))).decode()
        return DecryptedText

    @staticmethod
    def RSA_Encrypt(Text: str, p: int, q: int, e: int) -> str:
        """
        用于对字符串进行 RSA 加密。

        参数：
            Text (str): 待进行 RSA 加密的字符串
            p (int): p 值
            q (int): q 值
            e (int): e 值

        返回值：
            EncryptedText (str): RSA 加密后的密文（ Base64 编码形式）
        """

        m = bytes_to_long(Text.encode())
        c = gmpy2.powmod(m, e, p * q)
        EncryptedText = base64.b64encode(long_to_bytes(c)).decode()
        return EncryptedText

    @staticmethod
    def RSA_Base64_Decrypt(Base64Text: str, p: int, q: int, e: int) -> str:
        """
        用于对 Base64 编码形式的 RSA 密文进行解密。

        参数：
            Base64Text (str): 待进行解密的 Base64 编码形式的 RSA 密文
            p (int): p 值
            q (int): q 值
            e (int): e 值

        返回值：
            DecryptedText (str): RSA 解密后得到的原文
        """

        c = bytes_to_long(base64.b64decode(Base64Text.encode()))
        d = gmpy2.invert(e, (p - 1) * (q - 1))
        m = gmpy2.powmod(c, d, p * q)
        DecryptedText = long_to_bytes(m).decode()
        return DecryptedText

    @staticmethod
    def RSA_Long_Decrypt(Long: int, p: int, q: int, e: int) -> str:
        """
        用于对长整数形式的 RSA 密文进行解密。

        参数：
            Long (int): 待进行解密的长整数形式的 RSA 密文
            p (int): p 值
            q (int): q 值
            e (int): e 值

        返回值：
            DecryptedText (str): RSA 解密后得到的原文
        """

        d = gmpy2.invert(e, (p - 1) * (q - 1))
        m = gmpy2.powmod(Long, d, p * q)
        DecryptedText = long_to_bytes(m).decode()
        return DecryptedText

    @staticmethod
    def RSA_Wiener_Attack(c: int, e: int, n: int) -> str:
        """
        用于对长整数形式的 RSA 密文进行维纳攻击并解出原文。

        参数：
            c (int): 待进行维纳攻击的长整数形式的 RSA 密文
            e (int): e 值
            n (int): n 值

        返回值：
            m (str): RSA 维纳攻击后得到的原文
        """

        def continuedFra(x, y):
            """
            计算连分数

            参数：
                x: 分子
                y: 分母

            返回值：
                cf: 连分数列表
            """

            cf = []
            while y:
                cf.append(x // y)
                x, y = y, x % y
            return cf

        def gradualFra(cf):
            """
            计算连分数列表的最后的渐进分数

            参数：
                cf: 连分数列表

            返回值：
                numerator, denominator: 该列表最后的渐近分数（这里的渐进分数分子分母要分开）
            """

            numerator = 0
            denominator = 1
            for x in cf[::-1]:
                numerator, denominator = denominator, x * denominator + numerator
            return numerator, denominator

        def solve_pq(a, b, c):
            """
            使用韦达定理解出 pq (x^2−(p+q)∗x+pq=0)

            参数：
                a: x^2的系数
                b: x 的系数
                c: pq

            返回值：
                p, q: p 和 q 的值
            """

            par = gmpy2.isqrt(b * b - 4 * a * c)
            return (-b + par) // (2 * a), (-b - par) // (2 * a)

        def getGradualFra(cf):
            """
            计算连分数列表的所有的渐近分数

            参数：
                cf: 连分数列表

            返回值：
                gf: 该列表所有的渐近分数
            """

            gf = []
            for i in range(1, len(cf) + 1):
                gf.append(gradualFra(cf[:i]))
            return gf

        cf = continuedFra(e, n)
        gf = getGradualFra(cf)
        for d, k in gf:
            if k == 0:
                continue
            if (e * d - 1) % k != 0:
                continue
            phi = (e * d - 1) // k
            p, q = solve_pq(1, n - phi + 1, n)
            if p * q == n:
                m = long_to_bytes(gmpy2.powmod(c, d, n)).decode()
                return m

    @staticmethod
    def RSA_MultiPrime_Attack(c: int, e: int, n: int, primes: List[int], powers: Optional[List[int]] = None) -> str:
        """
        用于对长整数形式的 RSA 密文进行多素数攻击并解出原文。

        参数：
            c (int): 待进行多素数攻击的长整数形式的 RSA 密文
            e (int): e 值
            n (int): n 值
            primes (List[int]): 用于攻击的多素数列表
            powers (Optional[List[int]]): 各素数对应的阶数，默认均为 1 次方

        返回值：
            m (str): RSA 多素数攻击后得到的原文
        """

        from operator import mul
        from functools import reduce
        if powers == None:
            powers = [1 for _ in range(len(primes))]
        if len(primes) != len(powers):
            raise Exception("素数列表长度与阶数列表长度不一致")
        temp = [pow(primes[i], powers[i] - 1) * (primes[i] - 1) for i in range(len(primes))]
        phi = reduce(mul, temp)
        d = gmpy2.invert(e, phi)
        m = long_to_bytes(gmpy2.powmod(c, d, n)).decode()
        return m

    @staticmethod
    def RSA_LowEncryptionIndex_Attack(c: int, e: int, n: int) -> str:
        """
        用于对长整数形式的 RSA 密文进行低加密指数攻击并解出原文（尝试 10 万次累加 n 超过后会抛出异常）。

        参数：
            c (int): 待进行低加密指数攻击的长整数形式的 RSA 密文
            e (int): e 值
            n (int): n 值

        返回值：
            m (str): RSA 低加密指数攻击后得到的原文
        """

        k = 0
        while k < 10**5:
            m = gmpy2.iroot(c + k * n, e)[0]
            try:
                m = long_to_bytes(m).decode()
                return m
            except:
                k += 1
                continue
        raise Exception("尝试次数过多，攻击失败")

    @staticmethod
    def RSA_CommonMod_Attack(c1: int, c2: int, e1: int, e2: int, n: int) -> str:
        """
        用于对长整数形式的 RSA 密文进行共模攻击并解出原文。

        参数：
            c1 (int): 待进行共模攻击的长整数形式的第一串 RSA 密文
            c2 (int): 待进行共模攻击的长整数形式的第二串 RSA 密文
            e1 (int): c1 的 e 值
            e2 (int): c2 的 e 值
            n (int): n 值

        返回值：
            m (str): RSA 共模攻击后得到的原文
        """

        _, x, y = gmpy2.gcdext(e1, e2)
        m = long_to_bytes((gmpy2.powmod(c1, x, n) * gmpy2.powmod(c2, y, n)) % n).decode()
        return m

    @staticmethod
    def RSA_Broadcast_Attack(cs: List[int], e: int, ns: List[int]) -> str:
        """
        用于对长整数形式的 RSA 密文列表进行广播攻击并解出原文。

        参数：
            cs (List[int]): 待进行广播攻击的长整数形式的 RSA 密文列表
            e (int): e 值
            ns (List[int]): 各密文对应的 n 值的列表

        返回值：
            m (str): RSA 广播攻击后得到的原文
        """

        from functools import reduce
        if len(cs) != len(ns):
            raise Exception("密文列表长度与 n 值列表长度不一致")
        s = 0
        prod = reduce(lambda a, b: a * b, ns)
        for ni, ai in zip(ns, cs):
            p = prod // ni
            s += ai * gmpy2.invert(p, ni) * p
        x = s % prod
        m = long_to_bytes(gmpy2.iroot(x, e)[0]).decode()
        return m

    @staticmethod
    def RC4_Encrypt(Text: str, Key: str) -> str:
        """
        用于对字符串进行 RC4 加密。

        参数：
            Text (str): 待进行 RC4 加密的字符串
            Key (str): 加密密钥

        返回值：
            EncryptedText (str): RC4 加密后得到的密文（ Base64 编码形式）
        """

        from Crypto.Cipher import ARC4
        RC4 = ARC4.new(Key.encode())
        EncryptedText = base64.b64encode(RC4.encrypt(Text.encode())).decode()
        return EncryptedText

    @staticmethod
    def RC4_Decrypt(Text: str, Key: str) -> str:
        """
        用于对 Base64 编码形式的 RC4 密文进行解密。

        参数：
            Text (str): 待解密的 Base64 编码形式的 RC4 密文
            Key (str): 解密密钥

        返回值：
            DecryptedText (str): RC4 解密后得到的原文
        """

        from Crypto.Cipher import ARC4
        RC4 = ARC4.new(Key.encode())
        DecryptedText = RC4.decrypt(base64.b64decode(Text.encode())).decode()
        return DecryptedText


class ClassicalCryptoUtils():
    """
    本模块用于解决古典密码学问题。
    """

    @staticmethod
    def Caesar_Encrypt(Text: str, Move: int = 3) -> str:
        """
        用于对字符串进行 Caesar 加密。

        参数：
            Text (str): 待进行 Caesar 加密的字符串
            Move (可选)(int): 移位位数，默认为 3 

        返回值：
            EncryptedText (str): Caesar 加密后得到的密文
        """

        EncryptedText = ""
        for i in Text:
            if i.isupper():
                EncryptedText += chr((ord(i) - ord('A') + int(Move)) % 26 + ord('A'))
            elif i.islower():
                EncryptedText += chr((ord(i) - ord('a') + int(Move)) % 26 + ord('a'))
            else:
                EncryptedText += i
        return EncryptedText

    @staticmethod
    def Caesar_Decrypt(Text: str, Move: int = 3) -> str:
        """
        用于对 Caesar 密文进行解密。

        参数：
            Text (str): 待进行解密的 Caesar 密文
            Move (可选)(int): 移位位数，默认为 3 

        返回值：
            DecryptedText (str): Caesar 解密后得到的原文
        """

        DecryptedText = ""
        for i in Text:
            if i.isupper():
                DecryptedText += chr((ord(i) - ord('A') - int(Move)) % 26 + ord('A'))
            elif i.islower():
                DecryptedText += chr((ord(i) - ord('a') - int(Move)) % 26 + ord('a'))
            else:
                DecryptedText += i
        return DecryptedText

    @staticmethod
    def Caesar_Attack(Text: str) -> List[str]:
        """
        用于对 Caesar 密文进行爆破攻击。

        参数：
            Text (str): 待进行爆破攻击的 Caesar 密文

        返回值：
            Result (List[str]): Caesar 爆破攻击后得到的字符串列表
        """

        Result = [ClassicalCryptoUtils.Caesar_Decrypt(Text, i) for i in range(1, 27)]
        return Result

    @staticmethod
    def Morse_Encrypt(Text: str) -> str:
        """
        用于对字符串进行 Morse 加密。

        参数：
            Text (str): 待进行 Morse 加密的字符串

        返回值：
            EncryptedText (str): Morse 加密后得到的密文（未找到映射关系的字符将保持不变）
        """

        MorseCode = {
            "A": ".-", "B": "-...", "C": "-.-.", "D": "-..", "E": ".", "F": "..-.", "G": "--.",
            "H": "....", "I": "..", "J": ".---", "K": "-.-", "L": ".-..", "M": "--",
            "N": "-.", "O": "---", "P": ".--.", "Q": "--.-", "R": ".-.", "S": "...",
            "T": "-", "U": "..-", "V": "...-", "W": ".--", "X": "-..-", "Y": "-.--",
            "Z": "--..", "a": ".-", "b": "-...", "c": "-.-.", "d": "-..", "e": ".",
            "f": "..-.", "g": "--.", "h": "....", "i": "..", "j": ".---", "k": "-.-",
            "l": ".-..", "m": "--", "n": "-.", "o": "---", "p": ".--.", "q": "--.-",
            "r": ".-.", "s": "...", "t": "-", "u": "..-", "v": "...-", "w": ".--",
            "x": "-..-", "y": "-.--", "z": "--..", "0": "-----", "1": ".----", "2": "..---",
            "3": "...--", "4": "....-", "5": ".....", "6": "-....", "7": "--...", "8": "---..",
            "9": "----.", ".": ".-.-.-", ":": "---...", ",": "--..--", ";": "-.-.-.", "?": "..--..",
            "=": "-...-", "'": ".----.", "/": "-..-.", "!": "-.-.--", "-": "-....-", "_": "..--.-",
            '"': ".-..-.", "(": "-.--.", ")": "-.--.-", "$": "...-..-", "&": ".-...", "@": ".--.-.", "+": ".-.-."
        }
        EncryptedText = "/".join([MorseCode.get(i, i) for i in Text])
        return EncryptedText

    @staticmethod
    def Morse_Decrypt(Text: str) -> str:
        """
        用于对 Morse 密文进行解密。

        参数：
            Text (str): 待进行解密的 Morse 密文（以'/'进行分隔）

        返回值：
            DecryptedText (str): Morse 解密后得到的原文（未找到映射关系的字符将保持不变）
        """

        MorseCode = {
            ".-": "a", "-...": "b", "-.-.": "c", "-..": "d", ".": "e", "..-.": "f",
            "--.": "g", "....": "h", "..": "i", ".---": "j", "-.-": "k", ".-..": "l",
            "--": "m", "-.": "n", "---": "o", ".--.": "p", "--.-": "q", ".-.": "r",
            "...": "s", "-": "t", "..-": "u", "...-": "v", ".--": "w", "-..-": "x",
            "-.--": "y", "--..": "z", "-----": "0", ".----": "1", "..---": "2", "...--": "3",
            "....-": "4", ".....": "5", "-....": "6", "--...": "7", "---..": "8", "----.": "9",
            ".-.-.-": ".", "---...": ":", "--..--": ",", "-.-.-.": ";", "..--..": "?", "-...-": "=",
            ".----.": "'", "-..-.": "/", "-.-.--": "!", "-....-": "-", "..--.-": "_", ".-..-.": '"',
            "-.--.": "(", "-.--.-": ")", "...-..-": "$", ".-...": "&", ".--.-.": "@", ".-.-.": "+"
        }
        DecryptedText = "".join([MorseCode.get(i, "[" + i + "]") for i in Text.split("/")])
        return DecryptedText

    @staticmethod
    def Bacon_Encrypt(Text: str) -> str:
        """
        用于对字符串进行 Bacon 加密。

        参数：
            Text (str): 待进行 Bacon 加密的字符串

        返回值：
            EncryptedText (str): Bacon 加密后得到的密文（大写形式 未找到映射关系的字符将以[]包裹）
        """

        BaconCode = {
            "a": "aaaaa", "b": "aaaab", "c": "aaaba", "d": "aaabb", "e": "aabaa", "f": "aabab", "g": "aabba",
            "h": "aabbb", "i": "abaaa", "j": "abaab", "k": "ababa", "l": "ababb", "m": "abbaa", "n": "abbab",
            "o": "abbba", "p": "abbbb", "q": "baaaa", "r": "baaab", "s": "baaba", "t": "baabb",
            "u": "babaa", "v": "babab", "w": "babba", "x": "babbb", "y": "bbaaa", "z": "bbaab",
        }
        EncryptedText = "".join([BaconCode.get(i, "[" + i + "]") for i in Text.lower()]).upper()
        return EncryptedText

    @staticmethod
    def Bacon_Decrypt(Text: str) -> str:
        """
        用于对 Bacon 密文进行解密。

        参数：
            Text (str): 待进行解密的 Bacon 密文

        返回值：
            DecryptedText (str): Bacon 解密后得到的原文（大写形式 未找到映射关系的字符将以[]包裹）
        """

        if len(Text) % 5 != 0:
            raise Exception("Bacon 密文应该为 5 的倍数")
        BaconCode = {
            "aaaaa": "a", "aaaab": "b", "aaaba": "c", "aaabb": "d", "aabaa": "e", "aabab": "f", "aabba": "g",
            "aabbb": "h", "abaaa": "i", "abaab": "j", "ababa": "k", "ababb": "l", "abbaa": "m", "abbab": "n",
            "abbba": "o", "abbbb": "p", "baaaa": "q", "baaab": "r", "baaba": "s", "baabb": "t",
            "babaa": "u", "babab": "v", "babba": "w", "babbb": "x", "bbaaa": "y", "bbaab": "z",
        }
        List = [Text.lower()[i:i + 5] for i in range(0, len(Text), 5)]
        DecryptText = "".join([BaconCode.get(i, "[" + i + "]") for i in List]).upper()
        return DecryptText

    @staticmethod
    def Fence_Encrypt(Text: str, Fence: int) -> str:
        """
        用于对字符串进行 Fence 加密。

        参数：
            Text (str): 待进行 Fence 加密的字符串
            Fence (int): 栏数

        返回值：
            EncryptedText (str): Fence 加密后得到的密文
        """

        if len(Text) % Fence != 0:
            raise Exception("字符串长度应该为栏数的倍数")
        templist = []
        List = [Text[i:i + Fence] for i in range(0, len(Text), Fence)]
        for i in range(Fence):
            temp = ""
            for j in List:
                temp += j[i]
            templist.append(temp)
        EncryptedText = "".join(templist)
        return EncryptedText

    @staticmethod
    def Fence_Decrypt(Text: str, Fence: int) -> str:
        """
        用于对 Fence 密文进行解密。

        参数：
            Text (str): 待进行解密的 Fence 密文
            Fence (int): 栏数

        返回值：
            DecryptedText (str): Fence 解密后得到的原文
        """

        if len(Text) % Fence != 0:
            raise Exception("字符串长度应该为栏数的倍数")
        templist = ["" for _ in range(Fence)]
        List = [Text[i:i + Fence] for i in range(0, len(Text), Fence)]
        for i in List:
            for j in range(Fence):
                templist[j] = templist[j] + i[j]
        DecryptedText = "".join(templist)
        return DecryptedText

    @staticmethod
    def Fence_Attack(Text: str) -> List[tuple]:
        """
        用于对 Fence 密文进行爆破攻击。

        参数：
            Text (str): 待进行爆破攻击的 Fence 密文

        返回值：
            Result (List[tuple]): Fence 爆破攻击后得到的元组列表（字符串, 栏数）
        """

        Factors = [factor for factor in range(2, len(Text)) if len(Text) % factor == 0]
        Result = [(ClassicalCryptoUtils.Fence_Decrypt(Text, i), i) for i in Factors]
        return Result

    @staticmethod
    def WFence_Generate(Text: str, Fence: int) -> List[List[str]]:
        """
        用于生成 WFence 矩阵以便后续处理。

        参数：
            Text (str): 待进行 WFence 处理的字符串
            Fence (int): 栏数

        返回值：
            Matrix (List[List[str]]): 生成的 WFence 矩阵
        """

        Matrix = [['.'] * len(Text) for _ in range(Fence)]
        Row = 0
        Up = False
        for Column in range(len(Text)):
            Matrix[Row][Column] = Text[Column]
            if Row == Fence - 1:
                Up = True
            if Row == 0:
                Up = False
            if Up:
                Row -= 1
            else:
                Row += 1
        return Matrix

    @staticmethod
    def WFence_Encrypt(Text: str, Fence: int) -> str:
        """
        用于对字符串进行 WFence 加密。

        参数：
            Text (str): 待进行 WFence 加密的字符串
            Fence (int): 栏数

        返回值：
            EncryptedText (str): WFence 加密后得到的密文
        """

        Matrix = ClassicalCryptoUtils.WFence_Generate(Text, Fence)
        EncryptedText = ""
        for Row in range(Fence):
            for Column in range(len(Text)):
                if Matrix[Row][Column] != '.':
                    EncryptedText += Matrix[Row][Column]
        return EncryptedText

    @staticmethod
    def WFence_Decrypt(Text: str, Fence: int) -> str:
        """
        用于对 WFence 密文进行解密。

        参数：
            Text (str): 待进行解密的 WFence 密文
            Fence (int): 栏数

        返回值：
            DecryptedText (str): WFence 解密后得到的原文
        """

        Matrix = ClassicalCryptoUtils.WFence_Generate(Text, Fence)
        index = 0
        for Row in range(Fence):
            for Column in range(len(Text)):
                if Matrix[Row][Column] != '.':
                    Matrix[Row][Column] = Text[index]
                    index += 1
        DecryptedText = ""
        for Column in range(len(Text)):
            for Row in range(Fence):
                if Matrix[Row][Column] != '.':
                    DecryptedText += Matrix[Row][Column]
        return DecryptedText

    @staticmethod
    def WFence_Attack(Text: str) -> List[tuple]:
        """
        用于对 WFence 密文进行爆破攻击。

        参数：
            Text (str): 待进行爆破攻击的 WFence 密文

        返回值：
            Result (List[tuple]): WFence 爆破攻击后得到的元组列表（字符串, 栏数）
        """

        Result = [(ClassicalCryptoUtils.WFence_Decrypt(Text, i), i) for i in range(2, len(Text))]
        return Result



from setuptools import setup
import os


def long_description():
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'README.md'), encoding='utf-8') as f:
        return f.read()


setup(
    name="poseidon-python",
    version="1.1.0",
    author="B1ue1nWh1te",
    author_email="b1ue1nwh1te@skiff.com",
    description="海神波塞冬工具对常用的链上交互操作进行了封装，使得开发者能够便捷地与任何以太坊同构链交互，主要用于在CTF比赛中攻克Blockchain方向的题目。",
    keywords=['POSEIDON', 'BLOCKCHAIN', 'EVM', 'CRYPTO', 'CTF'],
    long_description=long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/B1ue1nWh1te/Poseidon",
    license="GPL-3.0",
    packages=["Poseidon"],
    python_requires='>=3.7',
    install_requires=["web3", "py-solc-x", "pyevmasm", "loguru", "pycryptodome", "gmpy2", "pwntools"],
    zip_safe=False
)



"""
本模块用于解决连接题目环境时可能遇到的工作量证明问题，在 Linux 环境下可以正常运行。
"""

import hashlib
from pwn import *
from pwnlib.util.iters import mbruteforce


class PoWUtils():
    """
    本模块用于解决连接题目环境时可能遇到的工作量证明问题，在 Linux 环境下可以正常运行。
    """

    @staticmethod
    def ProofOfWork_SHA256_Full(Url: str, Port: int, HashBegin: str, HashEnd: str, TextLength: int, SendAfter: str) -> remote:
        """
        用于解决连接题目环境时可能遇到的工作量证明问题，这一函数可处理以下情况：给出了 SHA256 的完整值，求长度为参数 (TextLength:int) 的字符串使得其 SHA256 值与给出的 SHA256 值相等。

        参数：
            Url (str): 题目环境的链接地址
            Port (int): 题目环境的端口号
            HashBegin (str): 哈希值之前的字符串
            HashEnd (str): 哈希值之后的字符串
            TextLength (int): 待求解的字符串的长度
            SendAfter (str): 在接收到这个参数所指明的字符串后才将求解出的字符串发送给服务器

        返回值：
            Connection (pwn.remote): 与服务器建立的连接对象
        """

        Connection = remote(Url, Port)
        Connection.recvuntil(HashBegin)
        Hash = Connection.recvuntil(HashEnd, drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: hashlib.sha256((x).encode()).hexdigest() == Hash, Charset, TextLength, method='fixed')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection

    @staticmethod
    def ProofOfWork_SHA256_Prefix(Url: str, Port: int, PrefixBegin: str, PrefixEnd: str, PrefixLength: int, MaxTextLength: int, SendAfter: str) -> remote:
        """
        用于解决连接题目环境时可能遇到的工作量证明问题，这一函数可处理以下情况：给出了 SHA256 的前缀值，求一个最大长度不超过参数 (MaxTextLength:int) 的字符串使得其 SHA256 值的前缀与给出的 SHA256 前缀相等。

        参数：
            Url (str): 题目环境的链接地址
            Port (int): 题目环境的端口号
            PrefixBegin (str): 哈希值前缀之前的字符串
            PrefixEnd (str): 哈希值前缀之后的字符串
            PrefixLength (int): 哈希值前缀的字符串的长度
            MaxTextLength (int): 待求解的字符串的最大长度
            SendAfter (str): 在接收到这个参数所指明的字符串后才将求解出的字符串发送给服务器

        返回值：
            Connection (pwn.remote): 与服务器建立的连接对象
        """

        Connection = remote(Url, Port)
        Connection.recvuntil(PrefixBegin)
        Prefix = Connection.recvuntil(PrefixEnd, drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: hashlib.sha256((x).encode()).hexdigest()[:PrefixLength] == Prefix, Charset, MaxTextLength, method='upto')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection

    @staticmethod
    def ProofOfWork_SHA256_EndWithZero(Url: str, Port: int, KnownBegin: str, KnownEnd: str, UnknownLength: int, EndWithZeroLength: int, SendAfter: str) -> remote:
        """
        用于解决连接题目环境时可能遇到的工作量证明问题，这一函数可处理以下情况：给出了原文的前一部分，剩下长度（UnknownLength:int）的部分未知，要求解出未知部分的字符串，以使得与已知部分拼接得到的字符串的 SHA256 值的二进制形式中末尾所包含的 0 的个数为（EndWithZeroLength:int）。

        参数：
            Url (str): 题目环境的链接地址
            Port (int): 题目环境的端口号
            KnownBegin (str): 已知部分之前的字符串
            KnownEnd (str): 已知部分之后的字符串
            UnknownLength (int): 未知部分的字符串的长度
            EndWithZeroLength (int): SHA256 值的二进制形式中末尾所包含的 0 的个数
            SendAfter (str): 在接收到这个参数所指明的字符串后才将求解出的字符串发送给服务器

        返回值：
            Connection (pwn.remote): 与服务器建立的连接对象
        """

        Connection = remote(Url, Port)
        Connection.recvuntil(KnownBegin)
        Known = Connection.recvuntil(KnownEnd, drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: bin(int(hashlib.sha256((Known + x).encode()).hexdigest(), 16)).endswith('0' * EndWithZeroLength), Charset, UnknownLength, method='fixed')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection

    @staticmethod
    def ProofOfWork_MD5_Full(Url: str, Port: int, HashBegin: str, HashEnd: str, TextLength: int, SendAfter: str) -> remote:
        """
        用于解决连接题目环境时可能遇到的工作量证明问题，这一函数可处理以下情况：给出了 MD5 的完整值，求长度为参数 (TextLength:int) 的字符串使得其 MD5 值与给出的 MD5 值相等。

        参数：
            Url (str): 题目环境的链接地址
            Port (int): 题目环境的端口号
            HashBegin (str): 哈希值之前的字符串
            HashEnd (str): 哈希值之后的字符串
            TextLength (int): 待求解的字符串的长度
            SendAfter (str): 在接收到这个参数所指明的字符串后才将求解出的字符串发送给服务器

        返回值：
            Connection (pwn.remote): 与服务器建立的连接对象
        """

        Connection = remote(Url, Port)
        Connection.recvuntil(HashBegin)
        Hash = Connection.recvuntil(HashEnd, drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: hashlib.md5((x).encode()).hexdigest() == Hash, Charset, TextLength, method='fixed')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection

    @staticmethod
    def ProofOfWork_MD5_Prefix(Url: str, Port: int, PrefixBegin: str, PrefixEnd: str, PrefixLength: int, MaxTextLength: int, SendAfter: str) -> remote:
        """
        用于解决连接题目环境时可能遇到的工作量证明问题，这一函数可处理以下情况：给出了 MD5 的前缀值，求一个最大长度不超过参数 (MaxTextLength:int) 的字符串使得其 MD5 值的前缀与给出的 MD5 前缀相等。

        参数：
            Url (str): 题目环境的链接地址
            Port (int): 题目环境的端口号
            PrefixBegin (str): 哈希值前缀之前的字符串
            PrefixEnd (str): 哈希值前缀之后的字符串
            PrefixLength (int): 哈希值前缀的字符串的长度
            MaxTextLength (int): 待求解的字符串的最大长度
            SendAfter (str): 在接收到这个参数所指明的字符串后才将求解出的字符串发送给服务器

        返回值：
            Connection (pwn.remote): 与服务器建立的连接对象
        """

        Connection = remote(Url, Port)
        Connection.recvuntil(PrefixBegin)
        Prefix = Connection.recvuntil(PrefixEnd, drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: hashlib.md5((x).encode()).hexdigest()[:PrefixLength] == Prefix, Charset, MaxTextLength, method='upto')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection

    @staticmethod
    def ProofOfWork_MD5_EndWithZero(Url: str, Port: int, KnownBegin: str, KnownEnd: str, UnknownLength: int, EndWithZeroLength: int, SendAfter: str) -> remote:
        """
        用于解决连接题目环境时可能遇到的工作量证明问题，这一函数可处理以下情况：给出了原文的前一部分，剩下长度（UnknownLength:int）的部分未知，要求解出未知部分的字符串，以使得与已知部分拼接得到的字符串的 MD5 值的二进制形式中末尾所包含的 0 的个数为（EndWithZeroLength:int）。

        参数：
            Url (str): 题目环境的链接地址
            Port (int): 题目环境的端口号
            KnownBegin (str): 已知部分之前的字符串
            KnownEnd (str): 已知部分之后的字符串
            UnknownLength (int): 未知部分的字符串的长度
            EndWithZeroLength (int): MD5 值的二进制形式中末尾所包含的 0 的个数
            SendAfter (str): 在接收到这个参数所指明的字符串后才将求解出的字符串发送给服务器

        返回值：
            Connection (pwn.remote): 与服务器建立的连接对象
        """

        Connection = remote(Url, Port)
        Connection.recvuntil(KnownBegin)
        Known = Connection.recvuntil(KnownEnd, drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: bin(int(hashlib.md5((Known + x).encode()).hexdigest(), 16)).endswith('0' * EndWithZeroLength), Charset, UnknownLength, method='fixed')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection
